import sys
from typing import List
from yakh.key._key import Key

try:

    import termios
    import tty
    import fcntl
    import os
    import re

    __ANSICODE = re.compile(r"\x1B[@-_][0-?]*[ -/]*[@-~]")

    __unconsumed_chars: List[str] = []

    # Adapted from: https://stackoverflow.com/a/59159112/9019559
    def __break_on_char(input_str: str) -> List[str]:
        pos = 0
        result = []
        for m in __ANSICODE.finditer(input_str):
            text = input_str[pos : m.start()]
            result.extend(list(text))
            result.append(m.group())
            pos = m.end()

        text = input_str[pos:]
        result.extend(list(text))
        return result

    def __get_key() -> str:
        fd_input = sys.stdin.fileno()
        term_attr = termios.tcgetattr(fd_input)
        fl = fcntl.fcntl(fd_input, fcntl.F_GETFL)
        fcntl.fcntl(fd_input, fcntl.F_SETFL, fl | os.O_NONBLOCK)
        tty.setraw(fd_input)
        ch_str = ""
        try:
            while True:
                if ch_str != "":
                    break
                while True:
                    ch_stri = sys.stdin.read(1)
                    if ch_stri == "":
                        break
                    ch_str += ch_stri
        except Exception:
            pass
        finally:
            termios.tcsetattr(fd_input, termios.TCSADRAIN, term_attr)
            fcntl.fcntl(fd_input, fcntl.F_SETFL, fl)
        first, *rest = __break_on_char(ch_str)
        __unconsumed_chars.extend(rest)
        return first

    def get_key() -> Key:
        """Returns a `Key` instance representing a keypress

        Returns:
            Key: Represents a keypress
        """
        if not __unconsumed_chars:
            ch_str = __get_key()
        else:
            ch_str = __unconsumed_chars.pop(0)

        ch_ord = tuple(map(ord, ch_str))
        return Key(ch_str, ch_ord, ch_str.isprintable() or ch_ord in [(13,), (27, 13)])

except ImportError:

    import msvcrt

    def get_key() -> Key:
        """Returns a `Key` instance representing a keypress

        Returns:
            Key: Represents a keypress
        """

        ch_str = msvcrt.getwch()
        if ch_str in ["à", "\x00"]:
            ch_str += msvcrt.getwch()
        ch_ord = tuple(map(ord, ch_str))
        return Key(ch_str, ch_ord, ch_str.isprintable())



from yakh._yakh import get_key
from yakh import key


"""
# yakh

yakh (Yet Another Keypress Handler) tries to handle keypresses from the stdin in the terminal in high-level platform indepdendent manner.

## Usage

```python
from yakh import get_key
from yakh.key import Keys

key = ''
while key not in ['q', Keys.ENTER]:
    key = get_key()
    if key.is_printable:
        print(key)
```

yakh is dead-simple, there is only one function `get_key()` which takes no arguments and blocks until a key is pressed.

For each keypress it creates an instance of [`Key`](./yakh/key/_key.py#L7) which holds:

- `.key`: characters representing the keypress
- `.key_codes`: collection of Unicode code point encodings for all the characters (given by `ord` function)
- `.is_printable`: printability of the characters in the keypress

Additionally `Key` instances

-  are comparable with another `Key` instances, `str` instances and *Unicode code point* representations (tuples of integers)
- come with string representation for purposes of printing and string concatenation, which returns the content of `.key` attribute

## `yakh.key` submodule
`yakh.key` sub-module contains platform dependent representations of certain keys under `Keys` class. These are available namely for `CTRL` key combinations and some other common keys. 

Full list of keys can be seen [here](./yakh/key/_key.py#L42) and [here](./yakh/key/_key.py#L81).
"""



from typing import Tuple, Union
from sys import platform

"""yakh.key` sub-module contains `Key` class which is used to represent keypresses and platform dependent representations of certain keys under `Keys` class. These are available namely for `CTRL` key combinations and some other common keys. 
"""


class Key:
    """`Key` class is used to represent keypresses

    Raises:
        ValueError: Raised when comparison is not possible
    """

    key: str
    key_codes: Tuple[int, ...]
    is_printable: bool

    def __init__(self, key: str, key_codes: Tuple[int, ...], is_printable: bool):
        self.key = key
        self.key_codes = key_codes
        self.is_printable = is_printable

    def __str__(self):
        return self.key

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other: Union[Tuple[int, ...], str]):
        if isinstance(other, str):
            return self.key == other

        if isinstance(other, tuple) and all(isinstance(item, int) for item in other):
            return self.key_codes == other

        if isinstance(other, Key):
            return all(
                [
                    self.key == other.key,
                    self.key_codes == other.key_codes,
                    self.is_printable == other.is_printable,
                ]
            )

        raise ValueError(f"Cannot compare with {other}. Expected `List[int]` or `str`.")

    def __ne__(self, other):
        return not self.__eq__(other)


class _PlatformIndependentKeys:
    CTRL_A = (1,)
    CTRL_B = (2,)
    CTRL_C = (3,)
    CTRL_D = (4,)
    CTRL_E = (5,)
    CTRL_F = (6,)
    CTRL_G = (7,)
    CTRL_H = (8,)
    CTRL_I = (9,)
    CTRL_J = (10,)
    CTRL_K = (11,)
    CTRL_L = (12,)
    CTRL_M = (13,)
    CTRL_N = (14,)
    CTRL_O = (15,)
    CTRL_P = (16,)
    CTRL_Q = (17,)
    CTRL_R = (18,)
    CTRL_S = (19,)
    CTRL_T = (20,)
    CTRL_U = (21,)
    CTRL_V = (22,)
    CTRL_W = (23,)
    CTRL_X = (24,)
    CTRL_Y = (25,)
    CTRL_Z = (26,)

    ENTER = (13,)
    ESC = (27,)

    TAB = (9,)

    def __init__(self):
        raise RuntimeError("`Keys` class cannot be instatiated")


if platform.startswith(("linux", "darwin", "freebsd")):

    class Keys(_PlatformIndependentKeys):
        UP_ARROW = (27, 91, 65)
        DOWN_ARROW = (27, 91, 66)
        RIGHT_ARROW = (27, 91, 67)
        LEFT_ARROW = (27, 91, 68)

        NUMPAD_UP_ARROW = (27, 91, 65)
        NUMPAD_DOWN_ARROW = (27, 91, 66)
        NUMPAD_RIGHT_ARROW = (27, 91, 67)
        NUMPAD_LEFT_ARROW = (27, 91, 68)

        BACKSPACE = (127,)
        DELETE = (27, 91, 51, 126)

        HOME = (27, 91, 72)
        END = (27, 91, 70)

        OPTION_ENTER = (27, 13)
        CTRL_ENTER = OPTION_ENTER

elif platform in ("win32", "cygwin"):

    class Keys(_PlatformIndependentKeys):
        UP_ARROW = (224, 72)
        DOWN_ARROW = (224, 80)
        RIGHT_ARROW = (224, 77)
        LEFT_ARROW = (224, 75)

        NUMPAD_UP_ARROW = (0, 72)
        NUMPAD_DOWN_ARROW = (0, 80)
        NUMPAD_RIGHT_ARROW = (0, 77)
        NUMPAD_LEFT_ARROW = (0, 75)

        BACKSPACE = (8,)
        DELETE = (224, 83)

        HOME = (224, 71)
        END = (224, 79)

        CTRL_ENTER = (10,)
        OPTION_ENTER = CTRL_ENTER

else:
    raise NotImplementedError(f"Platform `{platform}` is not supported")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yakh', 'yakh.key']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'python-yakh',
    'version': '0.3.1',
    'description': 'Yet Another Keypress Handler',
    'long_description': "# yakh\n\nyakh (Yet Another Keypress Handler) tries to handle keypresses from the stdin in the terminal in high-level platform indepdendent manner.\n\n## Installation\n\nYakh can be installed from [PyPI](https://pypi.org/project/python-yakh/) using,\n```sh\npip install python-yakh\n```\n\nor GitHub itself using poetry,\n```sh\npoetry add git+https://github.com/petereon/yakh.git\n```\n\n## Usage\n\n```python\nfrom yakh import get_key\nfrom yakh.key import Keys\n\nkey = ''\nwhile key not in ['q', Keys.ENTER]:\n    key = get_key()\n    if key.is_printable:\n        print(key)\n```\n\nyakh is dead-simple, there is only one function `get_key()` which takes no arguments and blocks until a key is pressed.\n\nFor each keypress it creates an instance of [`Key`](./yakh/key/_key.py#L7) which holds:\n\n- `.key`: characters representing the keypress\n- `.key_codes`: collection of Unicode code point encodings for all the characters (given by `ord` function)\n- `.is_printable`: printability of the characters in the keypress\n\nAdditionally `Key` instances\n\n-  are comparable with another `Key` instances, `str` instances and *Unicode code point* representations (tuples of integers)\n- come with string representation for purposes of printing and string concatenation, which returns the content of `.key` attribute\n\n## `yakh.key` submodule\n`yakh.key` sub-module contains platform dependent representations of certain keys under `Keys` class. These are available namely for `CTRL` key combinations and some other common keys. \n\nFull list of keys can be seen [here](./yakh/key/_key.py#L42) and [here](./yakh/key/_key.py#L81).\n",
    'author': 'petereon',
    'author_email': 'pvyboch1@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/petereon/yakh',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7.8,<4.0.0',
}


setup(**setup_kwargs)
